/* 
 *
 * closefd: closes a file descriptor of any process
 *
 * >>>>>>>>>>>>>>>>>>> FOR USE IN 2.2.X KERNELS <<<<<<<<<<<<<<<<<<<<<<<<<<<<<
 *
 * Example: insmod closefd.o pid=<num. of process> fd=<num. of file descriptor>
 *
 * Notes: 
 *        - Always returns EBUSY, there is nothing to worry about, this
 *          allows automatically unload
 *        - I'ts SMP safe
 *
 *****************************************************************************
 *
 *   Copyright (C) 2000 Ulisses Alonso Camar
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 *****************************************************************************
 *
 *   Please send me your comments, suggestions, job offerings/whatever to: 
 *   
 *                             uaca@alumni.uv.es
 */

#define MODULE /* si no... hacerlo con un /dev/? */

#include <linux/sched.h>

#ifdef MODULE

static int pid= -1;
static unsigned int fd=-1;

#include <linux/module.h>

MODULE_PARM(pid,"i");
MODULE_PARM(fd,"i"); /* --FIXME?: ? deberia ser unsigned */

#endif

struct task_struct *find_by_pid(int pid)
{
  struct task_struct *p, *init= (struct task_struct *)&init_task_union;

  for (p= init; (p= p->next_task)!= init && p->pid < pid; ) {
#ifdef __DEBUG__
    printk(KERN_INFO " pid %d: %s\n", p->pid, p->comm);
#endif
  }

  if (p->pid == pid)
    return p;
  else return NULL;
}

int closefd(int pid, unsigned int fd)
{
  struct task_struct *p;
  int error= -EINVAL;
  struct file * filp= NULL;
  
  if ((p= find_by_pid(pid)) == NULL) {
    return error;
  } 
  
  error = -EBADF;
  
  if (p->files && fd < p->files->max_fds)   /* filp = fcheck(fd); */
    filp = p->files->fd[fd];
  
  if (filp) {
    struct files_struct * files = p->files;
    files->fd[fd] = NULL;

    FD_CLR(fd, p->files->open_fds);     /* put_unused_fd(fd); */
    if (fd < p->files->next_fd)
      p->files->next_fd = fd;

    FD_CLR(fd, files->close_on_exec);
    error = filp_close(filp, files);
  }
  
  return error;
  
}

#ifdef MODULE

int init_module(void)
{
  int retval= -EINVAL;

  /*
  printk(KERN_INFO "Loading module closefd\n");
  */

  if (pid == -1 || fd == -1) {

    printk(KERN_INFO "Both pid and fd arguments are mandatory\n");
    return retval;
  }

  retval= closefd(pid, fd);
  
  printk(KERN_INFO "closefd(%i, %i) = %i\n", pid, fd, retval);

  return 1; /* Forzamos la descarga inmediatamente */
}

/*
void cleanup_module(void)
{
  printk(KERN_INFO "Unloading module of closefd\n");
}
*/

#endif

